@@ -37,55 +37,35 @@ import runpy
 # be *very* careful about what is imported here, as any stdlib loggers that are
 # created can not be changed!
 from talisker.config import get_config
-from talisker.util import ensure_extra_versions_supported, pkg_is_installed
-from talisker.context import CONTEXT, clear as clear_context  # noqa
+from talisker.util import (
+    ensure_extra_versions_supported,
+    pkg_is_installed,
+    flush_early_logs
+)
+from talisker.context import (  # NOQA
+    Context,
+    DeadlineExceeded,
+    request_timeout,
+)
 
-__version__ = '0.14.3'
+__version__ = '0.16.0'
 __all__ = [
     'initialise',
     'get_config',
-    'run',
-    'run_gunicorn',
-    'run_celery',
-    'run_gunicorn_eventlet',
-    'run_gunicorn_gevent',
+    'Context',
+    'DeadlineExceeded',
+    'request_timeout',
 ]
-_early_log_messages = []
 prometheus_multiproc_cleanup = False
 
 
-def _log(name, level, *args, **kwargs):
-    """Deferred log wrapper"""
-    logger = logging.getLogger(name)
-    getattr(logger, level)(*args, **kwargs)
-
-
-def early_log(name, level, *args, **kwargs):
-    """Logger wrap for talisker startup code.
-
-    Collects logs for later processing when logging is initialised
-    """
-    _early_log_messages.append((name, level, args, kwargs))
-
-
-def _flush_early_logs():
-    global early_log
-    # process pending logs
-    for name, level, args, kwargs in _early_log_messages:
-        _log(name, level, *args, **kwargs)
-    _early_log_messages[:] = []
-
-    # switch to immediate logging for any further early logs
-    early_log = _log
-
-
 def initialise(env=os.environ):
     global early_log
 
     config = get_config(env)
     import talisker.logs
     talisker.logs.configure(config)
-    _flush_early_logs()
+    flush_early_logs()
 
     # now that logging is set up, initialise other modules
     # sentry first, so we can report any further errors in initialisation
@@ -95,17 +75,21 @@ def initialise(env=os.environ):
         talisker.sentry.get_client()
     import talisker.statsd
     talisker.statsd.get_client()
-    clear_contexts()
+    clear_context()
     return config
 
 
-def clear_contexts():
+def clear_context():
     """Helper to clear any thread local contexts."""
     import talisker.sentry
-    clear_context()
+    Context.clear()
     talisker.sentry.clear()
 
 
+# b/w compat api
+clear_contexts = clear_context
+
+
 class RunException(Exception):
     pass
 
@@ -134,7 +118,7 @@ def run():
         sys.argv = sys.argv[1:]
         globs = {'__file__': script}
 
-        clear_contexts()
+        clear_context()
         runpy.run_path(script, globs, '__main__')
 
     except Exception:
@@ -220,19 +204,20 @@ def run_celery(argv=sys.argv):
     import talisker.celery
     from celery.bin.celery import main
     talisker.celery.enable_signals()
-    clear_contexts()
+    clear_context()
     main(argv)
 
 
 def run_gunicorn():
     config = get_config()
 
+    # Early throw-away parsing of gunicorn config, as we need to decide
+    # whether to enable prometheus multiprocess before we start importing
+    from gunicorn.app.wsgiapp import WSGIApplication
+    g_cfg = WSGIApplication().cfg
+
     # configure prometheus_client early as possible
     if pkg_is_installed('prometheus-client'):
-        # Early throw-away parsing of gunicorn config, as we need to decide
-        # whether to enable prometheus multiprocess before we start importing
-        from gunicorn.app.wsgiapp import WSGIApplication
-        g_cfg = WSGIApplication().cfg
         if g_cfg.workers > 1 or 'prometheus_multiproc_dir' in os.environ:
             from talisker.prometheus import setup_prometheus_multiproc
             async_workers = ('gevent', 'eventlet')
@@ -240,6 +225,23 @@ def run_gunicorn():
             setup_prometheus_multiproc(
                 any(n in g_cfg.worker_class_str for n in async_workers)
             )
+    try:
+        from gunicorn.workers.ggevent import GeventWorker
+        from talisker.context import enable_gevent_context
+    except Exception:
+        pass
+    else:
+        if g_cfg.worker_class == GeventWorker:
+            enable_gevent_context()
+
+    try:
+        from gunicorn.workers.geventlet import EventletWorker
+        from talisker.context import enable_eventlet_context
+    except Exception:
+        pass
+    else:
+        if g_cfg.worker_class == EventletWorker:
+            enable_eventlet_context()
 
     initialise()
 
@@ -251,7 +253,7 @@ def run_gunicorn():
 
     app = talisker.gunicorn.TaliskerApplication(
         "%(prog)s [OPTIONS] [APP_MODULE]", config.devel, config.debuglog)
-    clear_contexts()
+    clear_context()
     return app.run()
 
 
