@@ -30,13 +30,14 @@ from __future__ import absolute_import
 from builtins import *  # noqa
 
 from collections import OrderedDict
+from contextlib import contextmanager
 import logging
 import logging.handlers
 import numbers
 import sys
 import time
 
-from talisker.context import ContextStack, track_request_metric
+from talisker.context import Context, ContextId
 from talisker.util import (
     get_errno_fields,
     get_rounded_ms,
@@ -75,10 +76,25 @@ NOISY_LOGS = {
 }
 
 
-logging_context = ContextStack('logging')
+class LoggingContextProxy():
+
+    def __getattr__(self, attr):
+        return getattr(Context.logging, attr)
+
+    @contextmanager
+    def __call__(self, extra=None, **kwargs):
+        with Context.logging(extra, **kwargs):
+            yield
+
+
+logging_context = LoggingContextProxy()
+
+
+# backwards compat aliases
+def set_logging_context(*args, **kwargs):
+    Context.logging.push(*args, **kwargs)
+
 
-# backwards compat alias
-set_logging_context = logging_context.push
 extra_logging = logging_context
 
 
@@ -226,7 +242,7 @@ class StructuredLogger(logging.Logger):
     def _log(self, *args, **kwargs):
         t = time.time()
         super()._log(*args, **kwargs)
-        track_request_metric('logging', get_rounded_ms(t))
+        Context.track('logging', get_rounded_ms(t))
 
     # sadly, we must subclass and override, rather that use the new
     # setLogRecordFactory() in 3.2+, as that does not pass the extra args
@@ -246,7 +262,13 @@ class StructuredLogger(logging.Logger):
         structured = OrderedDict()
 
         try:
-            context_extra = logging_context.flat
+            if ContextId.get(None) is None:
+                context_extra = {}
+                request_id = None
+            else:
+                context_extra = logging_context.flat
+                request_id = Context.request_id
+
             global_extra = logging_globals.get('extra', {})
 
             if extra:
@@ -262,6 +284,8 @@ class StructuredLogger(logging.Logger):
                 structured[k] = v
 
             structured.update(global_extra)
+            if request_id:
+                structured['request_id'] = request_id
         except Exception:
             # ensure unexpected error doesn't break logging completely
             structured = extra
@@ -318,7 +342,6 @@ class StructuredFormatter(logging.Formatter):
             # we never want sentry to capture DEBUG logs in its breadcrumbs, as
             # they may be sensitive
             if record.levelno > logging.DEBUG:
-
                 talisker.sentry.record_log_breadcrumb(record)
 
             if len(record.message) > self.MAX_MSG_SIZE:
@@ -332,7 +355,7 @@ class StructuredFormatter(logging.Formatter):
             s = self._fmt % record.__dict__
 
             # add our structured tags *before* exception info is added
-            structured = getattr(record, '_structured', {})
+            structured = getattr(record, 'extra', {})
             if record.exc_info and 'errno' not in structured:
                 structured.update(get_errno_fields(record.exc_info[1]))
             if structured:
