@@ -30,19 +30,21 @@ from __future__ import absolute_import
 from builtins import *  # noqa
 __metaclass__ = type
 
+
 from collections import OrderedDict
+from datetime import datetime
 import logging
 import os
 import time
 import traceback
 import sys
+import uuid
 
-import talisker.context
+from talisker.context import Context
 import talisker.endpoints
 import talisker.requests
-import talisker.request_id
 import talisker.statsd
-from talisker.util import set_wsgi_header
+from talisker.util import set_wsgi_header, datetime_to_timestamp
 from talisker.render import (
     Content,
     Table,
@@ -63,8 +65,15 @@ __all__ = [
 REQUESTS = {}
 
 
+class RequestTimeout(Exception):
+    pass
+
+
 def talisker_error_response(environ, headers, exc_info):
-    """Returns WSGI iterable to be returned as an error response
+    """Returns WSGI iterable to be returned as an error response.
+
+    This error response uses Talisker's built in rendering support to be able
+    to render content in json (default), html, or text.
 
     Returns a tuple of (content_type, iterable)."""
     exc_type, exc, tb = exc_info
@@ -73,9 +82,6 @@ def talisker_error_response(environ, headers, exc_info):
 
     rid = environ['REQUEST_ID']
     id_info = [('Request-Id', rid)]
-    sentry_id = environ.get('SENTRY_ID')
-    if sentry_id:
-        id_info.append(('Sentry-ID', sentry_id))
 
     wsgi_environ = []
     request_headers = []
@@ -87,11 +93,11 @@ def talisker_error_response(environ, headers, exc_info):
             wsgi_environ.append((k, v))
 
     if config.devel:
-        title = 'Request {}: {}'.format(rid, exc)
+        title = '{}: {}'.format(exc_type.__name__, exc)
         lines = traceback.format_exception(*exc_info)
         tb = PreformattedText(''.join(lines), id='traceback')
     else:
-        title = 'Request {}: {}'.format(rid, exc_type.__name__)
+        title = 'Server Error: {}'.format(exc_type.__name__)
 
     content = [
         Content(title, tag='h1', id='title'),
@@ -156,7 +162,7 @@ class WSGIMetric:
     )
 
 
-class WSGIResponse():
+class TaliskerWSGIRequest():
     """Container for WSGI request/response cycle.
 
     It provides a start_response function that adds some headers, and captures
@@ -167,12 +173,12 @@ class WSGIResponse():
     def __init__(self,
                  environ,
                  start_response,
-                 added_headers=None,
-                 soft_timeout=-1):
+                 added_headers=None):
+
+        # request metadata
         self.environ = environ
         self.original_start_response = start_response
         self.added_headers = added_headers
-        self.soft_timeout = soft_timeout
 
         # response metadata
         self.status = None
@@ -185,6 +191,8 @@ class WSGIResponse():
         self.closed = False
         self.start_response_called = False
         self.start_response_timestamp = None
+        self.duration = 0
+        self.timedout = False
 
     def start_response(self, status, headers, exc_info=None):
         """Adds response headers and stores response data.
@@ -209,12 +217,9 @@ class WSGIResponse():
                 self.environ['REQUEST_ID'],
             )
 
-        if 'SENTRY_ID' in self.environ:
-            set_wsgi_header(
-                headers,
-                'X-Sentry-ID',
-                self.environ['SENTRY_ID'],
-            )
+        # are we going to be sending a sentry report? If so, include header
+        if self.exc_info or Context.debug:
+            set_wsgi_header(headers, 'X-Sentry-Id', self.environ['SENTRY_ID'])
 
         self.status = status
         status_code, _, _ = status.partition(' ')
@@ -230,7 +235,7 @@ class WSGIResponse():
         )
         self.start_response_called = True
 
-    def wrap(self, response_iter):
+    def wrap_response(self, response_iter):
         """Transforms this instance into an iterator that wraps the response.
 
         Allows for error handling and tracking response size.
@@ -274,7 +279,7 @@ class WSGIResponse():
 
         """
         if self.iter is None:
-            raise Exception("WSGIResponse: iterator has not been set yet")
+            raise Exception("iterator has not been set yet")
         # We don't actually call the WSGI server's provided start_response
         # until we are ready to start iterating the content. This provides us
         # with more control over the response, and allows us to more cleanly
@@ -296,33 +301,34 @@ class WSGIResponse():
         # as also after the first iteration, to work with both models.
         if self.status is not None and not self.start_response_called:
             self.call_start_response()
-
         try:
-            chunk = next(self.iter)
-            # support lazy WSGI apps, as above
-            if not self.start_response_called:
-                self.call_start_response()
-        except (StopIteration, GeneratorExit):
-            # support lazy WSGI apps with no content
-            if not self.start_response_called:
-                self.call_start_response()
-            # not all middleware calls close, so ensure it's called.
-            # Note: this does slightly affect the measured response latency,
-            # which will not include time spent closing the client socket
+            try:
+                chunk = next(self.iter)
+                # support lazy WSGI apps, as above
+                if not self.start_response_called:
+                    self.call_start_response()
+            except (StopIteration, GeneratorExit):
+                # support lazy WSGI apps with no content
+                if not self.start_response_called:
+                    self.call_start_response()
+                raise
+            except Exception as e:
+                self.exc_info = sys.exc_info()
+                if isinstance(e, RequestTimeout):
+                    self.timedout = True
+                # switch to generating an error response
+                self.iter = iter(self.error(self.exc_info))
+                chunk = next(self.iter)
+            except SystemExit as e:
+                if e.code != 0:
+                    self.exc_info = sys.exc_info()
+                raise
+        except (Exception, SystemExit):
+            # If the above has raised, it means that the request is done,
+            # While WSGI servers will call .close() on the iterator, middleware
+            # that wraps the iterator may not, so we call close manually.
             self.close()
             raise
-        except Exception:
-            self.report_error()
-            # switch to generating an error response
-            self.iter = iter(self.error(sys.exc_info()))
-            chunk = next(self.iter)
-        except KeyboardInterrupt:
-            self.report_error()
-            raise
-        except SystemExit as e:
-            if e.code != 0:
-                self.report_error()
-            raise
 
         self.content_length += len(chunk)
         return chunk
@@ -363,50 +369,179 @@ class WSGIResponse():
             self.finish_request()
             self.closed = True
 
-    def finish_request(self):
+    def finish_request(self, timeout=False):
+        if timeout:
+            self.timedout = timeout
         start = self.environ.get('start_time')
-        duration = 0
         response_latency = 0
         if start:
-            duration = time.time() - start
-            response_latency = (self.start_response_timestamp - start) * 1000
+            self.duration = time.time() - start
+            if self.start_response_timestamp:
+                response_latency = (
+                    (self.start_response_timestamp - start) * 1000
+                )
 
-        log_response(
-            self.environ,
-            self.status_code,
-            self.headers,
-            duration,
-            self.content_length,
-            exc_info=self.exc_info,
-            filepath=self.file_path,
-        )
+        metadata = self.get_metadata()
+        self.log(metadata)
+        self.metrics(metadata)
 
-        if self.soft_timeout > 0 and response_latency > self.soft_timeout:
+        # We want to send a sentry report if:
+        # a) an error or timeout occured
+        # b) soft timeout
+        # c) manual debugging
+
+        if talisker.sentry.enabled:
+            soft_timeout = Context.current.soft_timeout
             try:
-                talisker.sentry.report_wsgi_error(
-                    self.environ,
-                    msg='Start_response over timeout: {}ms'
-                        .format(self.soft_timeout),
-                    level='warning')
+                if self.exc_info:
+                    self.send_sentry(metadata)
+                elif Context.debug:
+                    self.send_sentry(
+                        metadata,
+                        msg='Debug: {}'.format(metadata['path']),
+                        level='debug',
+                    )
+                elif soft_timeout > 0 and response_latency > soft_timeout:
+                    self.send_sentry(
+                        metadata,
+                        msg='Soft Timeout: {}'.format(metadata['path']),
+                        level='warning',
+                        extra={
+                            'start_response_latency': response_latency,
+                            'soft_timeout': soft_timeout,
+                        },
+                    )
             except Exception:
                 logger.exception('failed to send soft timeout report')
 
-        talisker.clear_contexts()
+        talisker.clear_context()
         rid = self.environ.get('REQUEST_ID')
         if rid:
             REQUESTS.pop(rid, None)
 
-    def report_error(self):
-        sentry_id = talisker.sentry.report_wsgi_error(self.environ)
-        if sentry_id is not None:
-            self.environ['SENTRY_ID'] = sentry_id
+    def get_metadata(self):
+        """Return an ordered dictionary of request metadata for logging."""
+        environ = self.environ
+        extra = OrderedDict()
+
+        if self.headers is None:
+            headers = {}
+        else:
+            headers = dict((k.lower(), v) for k, v in self.headers)
+
+        extra['method'] = environ.get('REQUEST_METHOD')
+        script = environ.get('SCRIPT_NAME', '')
+        path = environ.get('PATH_INFO', '')
+        extra['path'] = script + '/' + path.lstrip('/')
+        qs = environ.get('QUERY_STRING')
+        if qs:
+            extra['qs'] = environ.get('QUERY_STRING')
+        if self.status_code:
+            extra['status'] = self.status_code
+        if 'VIEW_NAME' in environ:
+            extra['view'] = environ['VIEW_NAME']
+        elif 'x-view-name' in headers:
+            extra['view'] = headers['x-view-name']
+        extra['duration_ms'] = round(self.duration * 1000, 3)
+        extra['ip'] = environ.get('REMOTE_ADDR', None)
+        extra['proto'] = environ.get('SERVER_PROTOCOL')
+        if self.content_length:
+            extra['length'] = self.content_length
+        if self.file_path is not None:
+            extra['filepath'] = self.file_path
+        request_length = environ.get('CONTENT_LENGTH')
+        if request_length:
+            try:
+                extra['request_length'] = int(request_length)
+            except ValueError:
+                pass
+        content_type = environ.get('CONTENT_TYPE')
+        if content_type:
+            extra['request_type'] = content_type
+        referrer = environ.get('HTTP_REFERER', None)
+        if referrer is not None:
+            extra['referrer'] = environ.get('HTTP_REFERER', None)
+        if 'HTTP_X_FORWARDED_FOR' in environ:
+            extra['forwarded'] = environ['HTTP_X_FORWARDED_FOR']
+        if 'HTTP_USER_AGENT' in environ:
+            extra['ua'] = environ['HTTP_USER_AGENT']
+
+        if self.timedout:
+            extra['timeout'] = True
+
+        if self.exc_info and self.exc_info[0]:
+            extra['exc_type'] = str(self.exc_info[0].__name__)
+            extra['trailer'] = ''.join(
+                traceback.format_exception(*self.exc_info)
+            )
+
+        tracking = Context.current.tracking
+        for name, tracker in sorted(tracking.items()):
+            extra[name + '_count'] = tracker.count
+            extra[name + '_time_ms'] = tracker.time
+
+        return extra
+
+    def log(self, extra):
+        """Log a WSGI request.
+
+        Similar to access logs, but structured and with more data."""
+        try:
+            msg = "{method} {path}{0}".format(
+                '?' if 'qs' in extra else '',
+                **extra
+            )
+        except Exception:
+            logger.exception('error generating access log', extra=extra)
+        else:
+            logger.info(msg, extra=extra)
+
+    def metrics(self, extra):
+        labels = {
+            'view': extra.get('view', 'unknown'),
+            'method': extra['method'],
+            'status': str(extra.get('status', 'timeout')).lower(),
+        }
+
+        WSGIMetric.requests.inc(**labels)
+        WSGIMetric.latency.observe(extra['duration_ms'], **labels)
+        if self.timedout:
+            lbls = labels.copy()
+            lbls.pop('status')
+            WSGIMetric.timeouts.inc(**lbls)
+        elif self.status_code and self.status_code >= 500:
+            WSGIMetric.errors.inc(**labels)
+
+    def send_sentry(self, metadata, msg=None, data=None, **kwargs):
+        from raven.utils.wsgi import get_current_url, get_environ, get_headers
+        if data is None:
+            data = {}
+        if 'SENTRY_ID' in self.environ:
+            data['event_id'] = self.environ['SENTRY_ID']
+        # sentry displays these specific fields in a different way
+        http_context = {
+            'url': get_current_url(self.environ),
+            # we don't use the sanitized version from metadata, we want the
+            # real query string
+            'query_string': self.environ.get('QUERY_STRING'),
+            'method': metadata['method'],
+            'headers': dict(get_headers(self.environ)),
+            'env': dict(get_environ(self.environ)),
+        }
+        talisker.sentry.report_wsgi(
+            http_context,
+            exc_info=self.exc_info,
+            msg=msg,
+            data=data,
+            **kwargs
+        )
 
 
 class TaliskerMiddleware():
     """Talisker entrypoint for WSGI apps.
 
     Sets up some values in environ, handles errors, and wraps responses in
-    WSGIResponse.
+    TaliskerWSGIRequest.
     """
     def __init__(self, app, environ=None, headers=None):
 
@@ -420,153 +555,103 @@ class TaliskerMiddleware():
         self.environ = environ
         self.headers = headers
 
+        # ensure new workers have an initialised sentry_context
+        talisker.sentry.new_context()
+
     def __call__(self, environ, start_response):
-        start_time = time.time()
+        Context.new()
         config = talisker.get_config()
-        environ['start_time'] = start_time
+
+        # setup environment
+        environ['start_time'] = Context.current.start_time
         if self.environ:
             environ.update(self.environ)
-
         # ensure request id
         if config.wsgi_id_header not in environ:
-            environ[config.wsgi_id_header] = talisker.request_id.generate()
+            environ[config.wsgi_id_header] = str(uuid.uuid4())
         rid = environ[config.wsgi_id_header]
         environ['REQUEST_ID'] = rid
-        talisker.request_id.push(rid)
+        # needs to be different from request id, as request can be passed on to
+        # upstream services
+        environ['SENTRY_ID'] = uuid.uuid4().hex
 
-        REQUESTS[rid] = environ
+        Context.request_id = rid
+        Context.current.soft_timeout = config.soft_request_timeout
 
-        response = WSGIResponse(
-            environ,
-            start_response,
-            self.headers,
-            config.soft_request_timeout,
-        )
+        # calculate ip route
+        route = None
+        try:
+            forwarded = environ.get('HTTP_X_FORWARDED_FOR')
+            if forwarded:
+                route = [a.strip() for a in forwarded.split(',')]
+            elif "REMOTE_ADDR" in environ:
+                route = [environ["REMOTE_ADDR"]]
+        except Exception as e:
+            logger.exception(e)
+        else:
+            if route is not None:
+                environ['ACCESS_ROUTE'] = route
+                environ['CLIENT_ADDR'] = route[-1]
+
+        if 'HTTP_X_DEBUG' in environ:
+            if config.is_trusted_addr(environ.get('CLIENT_ADDR')):
+                Context.set_debug()
+            else:
+                logger.warning(
+                    'X-Debug header set but not trusted IP address',
+                    extra={
+                        "access_route": ','.join(environ.get('ACCESS_ROUTE')),
+                        "x_debug": environ['HTTP_X_DEBUG'],
+                    }
+                )
+
+        set_deadline = False
+        header_deadline = environ.get('HTTP_X_REQUEST_DEADLINE')
+        if header_deadline:
+            try:
+                deadline = datetime.strptime(
+                    header_deadline,
+                    "%Y-%m-%dT%H:%M:%S.%fZ",
+                )
+            except ValueError:
+                pass
+            else:
+                # set deadline directly
+                # TODO: validate deadline is in future?
+                Context.current.deadline = datetime_to_timestamp(deadline)
+                set_deadline = True
+
+        if not set_deadline and config.request_timeout is not None:
+            Context.current.set_deadline(config.request_timeout)
+
+        # create the response container
+        request = TaliskerWSGIRequest(environ, start_response, self.headers)
+
+        # track in-flight requests
+        if rid in REQUESTS:
+            logger.warning(
+                'duplicate request id received by gunicorn worker',
+                extra={'request_id': rid}
+            )
+        else:
+            REQUESTS[rid] = request
 
         try:
-            response_iter = self.app(environ, response.start_response)
-        except Exception:
-            response.report_error()
-            response_iter = response.error(sys.exc_info())
-        except KeyboardInterrupt:
-            response.report_error()
-            raise
+            response_iter = self.app(environ, request.start_response)
+        except Exception as e:
+            # store details for later
+            request.exc_info = sys.exc_info()
+            if isinstance(e, RequestTimeout):
+                request.timedout = True
+            # switch to generating an error response
+            response_iter = request.error(request.exc_info)
         except SystemExit as e:
             if e.code != 0:
-                response.report_error()
+                request.exc_info = sys.exc_info()
+                request.finish_request()
             raise
 
-        return response.wrap(response_iter)
-
-
-def get_metadata(environ,
-                 status=None,
-                 headers=None,
-                 duration=None,
-                 length=None,
-                 exc_info=None,
-                 filepath=None):
-    """Return an ordered dictionary of request metadata for logging."""
-    if headers is None:
-        headers = {}
-    else:
-        headers = dict((k.lower(), v) for k, v in headers)
-    extra = OrderedDict()
-    extra['method'] = environ.get('REQUEST_METHOD')
-    script = environ.get('SCRIPT_NAME', '')
-    path = environ.get('PATH_INFO', '')
-    extra['path'] = script + '/' + path.lstrip('/')
-    qs = environ.get('QUERY_STRING')
-    if qs:
-        extra['qs'] = environ.get('QUERY_STRING')
-    if status:
-        extra['status'] = status
-    if 'VIEW_NAME' in environ:
-        extra['view'] = environ['VIEW_NAME']
-    elif 'x-view-name' in headers:
-        extra['view'] = headers['x-view-name']
-    if duration:
-        extra['duration_ms'] = round(duration * 1000, 3)
-    extra['ip'] = environ.get('REMOTE_ADDR', None)
-    extra['proto'] = environ.get('SERVER_PROTOCOL')
-    if length:
-        extra['length'] = length
-    if filepath is not None:
-        extra['filepath'] = filepath
-    request_length = environ.get('CONTENT_LENGTH')
-    if request_length:
-        try:
-            extra['request_length'] = int(request_length)
-        except ValueError:
-            pass
-    content_type = environ.get('CONTENT_TYPE')
-    if content_type:
-        extra['request_type'] = content_type
-    referrer = environ.get('HTTP_REFERER', None)
-    if referrer is not None:
-        extra['referrer'] = environ.get('HTTP_REFERER', None)
-    if 'HTTP_X_FORWARDED_FOR' in environ:
-        extra['forwarded'] = environ['HTTP_X_FORWARDED_FOR']
-    if 'HTTP_USER_AGENT' in environ:
-        extra['ua'] = environ['HTTP_USER_AGENT']
-
-    if exc_info and exc_info[0]:
-        extra['exc_type'] = str(exc_info[0].__name__)
-        extra['trailer'] = ''.join(
-            traceback.format_exception(*exc_info)
-        )
-
-    tracking = getattr(talisker.context.CONTEXT, 'request_tracking', {})
-    for name, tracker in tracking.items():
-        extra[name + '_count'] = tracker.count
-        extra[name + '_time_ms'] = tracker.time
-
-    msg = "{method} {path}{0}".format('?' if 'qs' in extra else '', **extra)
-    return msg, extra
-
-
-def log_response(environ,
-                 status=None,
-                 headers=None,
-                 duration=None,
-                 length=None,
-                 exc_info=None,
-                 filepath=None,
-                 timeout=False,
-                 **kwargs):
-    """Log a WSGI request and record metrics.
-
-    Similar to access logs, but structured and with more data."""
-    try:
-        msg, extra = get_metadata(
-            environ,
-            status,
-            headers,
-            duration,
-            length,
-            exc_info,
-            filepath,
-        )
-        if timeout:
-            extra['timeout'] = True
-        logger.info(msg, extra=extra)
-    except Exception:
-        logger.exception('error generating access log')
-    else:
-        labels = {
-            'view': extra.get('view', 'unknown'),
-            'method': extra['method'],
-            'status': str(status).lower(),
-        }
-
-        WSGIMetric.requests.inc(**labels)
-        WSGIMetric.latency.observe(extra['duration_ms'], **labels)
-        if status is None and timeout:
-            lbls = labels.copy()
-            lbls.pop('status')
-            WSGIMetric.timeouts.inc(**lbls)
-        elif status >= 500:
-            WSGIMetric.errors.inc(**labels)
+        return request.wrap_response(response_iter)
 
 
 def wrap(app):
